/*
 * @(#)ExternalCharsetProvider.java        0.0        09/06/03
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Cologne <Ulf.Zibis at CoSoCo.de>
 */

package sun.nio.cs;

import java.io.DataInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Base class for external charset providers.
 *
 * @author Ulf Zibis, Cologne <Ulf.Zibis at CoSoCo.de>
 */
public class ExternalCharsetProvider extends FastCharsetProvider {

    // Maps uppercased canonical names to class names
    private final Map<String,String> classMap;

    // Maps uppercased alias or canonical names to canonical names
    private final Map<String,String> canonicalMap;

    // Maps uppercased canonical names to alias-name arrays
    private final Map<String,String[]> aliasesMap;

    protected ExternalCharsetProvider(InitData initData) {
        this(initData.charsetsCount, initData.classNames, initData.canonicalNames, initData.aliases);
    }
    protected ExternalCharsetProvider(int length, Map<String,String> classNames,
            Map<String,String> uppercasedCanonicalNames, Map<String,String[]> aliases) {
        super(length, new HashMap());
        classMap = classNames;
        canonicalMap = uppercasedCanonicalNames;
        aliasesMap = aliases;
    }

    private static <T> T remove(Map<String,T> m, String uppercasedCanonicalName) {
        T x  = m.remove(uppercasedCanonicalName);
        assert (x != null);
        return x;
    }

    /** Declare support for the given charset, undeclare before registered namings accordingly */
    protected synchronized void register(String canonicalName, String className, String... aliases) {
        String uppercasedCanonical = toUpper(canonicalName);
            // registerd charset names are held in 4 locations:
            // (1) classMap
            // (2) canonicalMap
            // (3) aliasesMap
            // or if yet in use:
            // (4) cache, holds charsets by canonical name and aliases (yet unused remain in canonicalMap)
        // overwrite (1) and/or remove from cache (4) and check if already registered
        if (classMap.put(uppercasedCanonical, className) != null
                || cache.remove(uppercasedCanonical) != null)
            throw new IllegalCharsetNameException("Charset ["+canonicalName+"] is duplicate and yet registered");
        // register new aliases in (2), (3)
        if (aliases != null) {
            String oldCanonical;
            for (int i=0; i<aliases.length; i++)
                if ((oldCanonical = canonicalMap.put(toUpper(aliases[i]), canonicalName)) != null) {
                    for (int j=0; j<i; j++) // remove yet registered aliases
                        canonicalMap.remove(toUpper(aliases[j]));
                    remove(classMap, uppercasedCanonical);
                    canonicalMap.put(toUpper(aliases[i]), oldCanonical); // while already there, re-put it and ...
                    // should not happen, if data is generated from build.tools.charsetmapping.CharsetsDataGenerator
                    throw new IllegalCharsetNameException("Alias ["+aliases[i]+"] is duplicate and yet assigned to charset ["+oldCanonical+"]");
                }
            // overwrite (3)
            aliasesMap.put(uppercasedCanonical, aliases);
        }
        // overwrite (2)
        canonicalMap.put(uppercasedCanonical, canonicalName);
}

    /**
     * Undeclare support for the given charset and return it's before registered
     * aliase names, class name and canonical name in this order.
     */
    protected synchronized List<String> unregister(String charsetName) {
        charsetName = toUpper(charsetName);
        Charset cs = cache.remove(charsetName);
        String[] aliases = cs == null ? remove(aliasesMap, charsetName)
            : cs.aliases().toArray(new String[cs.aliases().size()]);
        // TODO: could be removed when all sun.nio.cs.ext charsets were updated accordingly:
            remove(aliasesMap, charsetName);
        List<String> names = new ArrayList();
        if (aliases != null) {
            for (String alias : aliases)
                if (cache.remove(alias = toUpper(alias)) == null)
                    remove(canonicalMap, alias);
            names.addAll(Arrays.asList(aliases));
        }
        if (cs == null) {
            names.add(remove(classMap, charsetName));
            if (names.get(0) == null)
                return null; // given charsetName was not registered
            names.add(remove(canonicalMap, charsetName));
        } else {
            String className = (cs instanceof FastCharset)
            // following className hack is temporary; fix for Bug 6862165 will change this
            // add indicator for generic class:
                    ? '\\'+((FastCharset)cs).mappingClsName
                    : cs.getClass().getSimpleName();
            names.add(className);
            names.add(cs.name());
        }
        return names;
    }

    protected synchronized Set<Charset> charsets0() {
        if (charsets.isEmpty()) {
            super.charsets();
            for (String uppercasedName : classMap.keySet())
                charsets.add(lookup(uppercasedName));
        }
        return charsets;
    }

    protected synchronized Charset lookup(String uppercasedName) {
        Charset cs = cache.get(uppercasedName);
        if (cs == null) {
            String canonicalName = canonicalMap.remove(uppercasedName); // no more needed
            if (canonicalName == null)
                return null; // name is unknown here
            String uppercasedCanonical = toUpper(canonicalName);
            String className = classMap.remove(uppercasedCanonical); // no more needed
            if (className != null) {
//                cs = newCharset(className, canonicalName, aliasesMap.remove(uppercasedCanonical)); // no more needed
                cs = newCharset(className, canonicalName, aliasesMap.get(uppercasedCanonical)); // workaround until legacy classes are removed
                assert cs != null;
                if (canonicalMap.remove(uppercasedCanonical) != null)
                    cache.put(uppercasedCanonical, cs);
            } else
                cs = cache.get(uppercasedCanonical);
            assert cs != null;
            cache.put(uppercasedName, cs);
        }
        return cs;
    }

    private Charset newCharset(String className, String canonicalName, String[] aliases) {
        // following className hack is temporary; fix for Bug 6862165 will change this
        if (className.charAt(0) == '\\')
            return new SoftReferenceCharset(packagePrefix, className.substring(1), canonicalName, aliases);
        try {
            try {
                return (Charset)Class.forName(packagePrefix+'.'+className, true, getClass().getClassLoader())
                        .getConstructor(String.class, String[].class).newInstance(canonicalName, aliases);
            } catch (NoSuchMethodException ex) { // workaround for legacy classes
                return (Charset)Class.forName(packagePrefix+'.'+className,
                        true, getClass().getClassLoader()).newInstance();
            }
//        } catch (ReflectiveOperationException ex) {
//            return null;
        } catch (ClassNotFoundException ex) {
//        } catch (NoSuchMethodException ex) { // needed when catch for legacy classes is removed
        } catch (IllegalArgumentException ex) {
        } catch (InvocationTargetException ex) {
        } catch (InstantiationException ex) {
        } catch (IllegalAccessException ex) {
        }
        return null;
    }

//    // TODO: could be removed when all sun.nio.cs.ext charsets were updated accordingly
//    protected String[] aliases(String uppercasedCanonical) {
//        return aliasesMap.get(uppercasedCanonical);
//    }

    protected static class InitData {
        private final int charsetsCount;
        private final int classesCount;
        private final int allNamesCount;
        private final Map<String,String> canonicalNames;
        private final Map<String,String> classNames;
        private final Map<String,String[]> aliases;
        public InitData(Class<? extends ExternalCharsetProvider> enclosing) throws IOException {
            DataInputStream in = new DataInputStream(enclosing.getResourceAsStream(enclosing.getSimpleName()+".dat"));
            charsetsCount = in.readShort();
            classesCount = in.readShort();
            allNamesCount = in.readShort();
            canonicalNames = new HashMap(allNamesCount, 1); // purged after usage,
            classNames = new HashMap(charsetsCount, 1);     // only needed for
            aliases = new HashMap(charsetsCount, 1);        // instantiating references
            for (int i=0; i<charsetsCount; i++) {
                String currentCanonical = in.readUTF().intern();
                String uppercasedCanonical = toUpper(currentCanonical);
                String oldCanonical = canonicalNames.put(uppercasedCanonical, currentCanonical);
                assert oldCanonical == null // should not happen, as yet catched from CharsetsDataGenerator
                        : "Canonical name ("+currentCanonical+") is duplicate of charset ("+oldCanonical+")";
                classNames.put(uppercasedCanonical, in.readUTF().intern());
                String[] aliases = new String[in.readByte()];
                for (int j=0; j<aliases.length; j++) {
                    aliases[j] = in.readUTF().intern();
                    oldCanonical = canonicalNames.put(toUpper(aliases[j]), currentCanonical);
                    assert oldCanonical == null // should not happen, as yet catched from CharsetsDataGenerator
                            : "Alias ("+aliases[j]+") is duplicate and yet assigned to charset ("+oldCanonical+")";
                }
                this.aliases.put(uppercasedCanonical, aliases);
            }
        }
    }
}
