/*
 * Copyright 2009 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package sun.nio.cs.ext;

import java.nio.CharBuffer;
import java.nio.ByteBuffer;
import java.nio.charset.CoderResult;
import java.util.Arrays;
import sun.nio.cs.SoftReferenceCharset;
import sun.nio.cs.FastCharset;
import static sun.nio.cs.Mapping.*;

/**
 *
 * @author unascribed
 * @author Ulf Zibis, Cologne <Ulf.Zibis @ CoSoCo.de>
 */
public class EUC_TW extends SoftReferenceCharset {

//    private static final int SS2 = 0x8E;
    private static final byte SS2 = (byte)0x8E;

    /*
       (1) EUC_TW
       Second byte of EUC_TW for cs2 is in range of
       0xA1-0xB0 for plane 1-16. According to CJKV /163,
       plane1 is coded in both cs1 and cs2. This impl
       however does not decode the codepoints of plane1
       in cs2, so only p2-p7 and p15 are supported in cs2.

       Plane2  0xA2;
       Plane3  0xA3;
       Plane4  0xA4;
       Plane5  0xA5;
       Plane6  0xA6;
       Plane7  0xA7;
       Plane15 0xAF;

       (2) Mapping
       The fact that all supplementary characters encoded in EUC_TW are
       in 0x2xxxx range gives us the room to optimize the data tables.

       Decoding:
       (1) save the lower 16-bit value of all codepoints of b->c mapping
           in a String array table  String[plane] b2c.
       (2) save "codepoint is supplementary" info (one bit) in a
           byte[] b2cIsSupp, so 8 codepoints (same codepoint value, different
           plane No) share one byte.

       Encoding:
       (1)c->b mappings are stored in
          char[]c2b/char[]c2bIndex
          char[]c2bSupp/char[]c2bIndexsupp  (indexed by lower 16-bit
       (2)byte[] c2bPlane stores the "plane info" of each euc-tw codepoints,
          BMP and Supp share the low/high 4 bits of one byte.

       Mapping tables are stored separated in EUC_TWMapping, which
       is generated by tool.
     */

    public EUC_TW(String canonicalName, String... aliases) {
        super(EUC_TWMapping.class.getPackage().getName(),
                EUC_TWMapping.class.getSimpleName(), canonicalName, aliases);
    }

    public static class Decoder extends SoftReferenceCharset.Decoder<EUC_TWMapping> {

        private final String[] b2c     = mapping.b2c;
//        private final int b1Min        = mapping.b1Min;
//        private final int b1Max        = mapping.b1Max;
//        private final int b2Min        = mapping.b2Min;
//        private final int b2Max        = mapping.b2Max;
        private final byte b1Min       = mapping.b1Min;
        private final byte b1Max       = mapping.b1Max;
        private final byte b2Min       = mapping.b2Min;
        private final byte b2Max       = mapping.b2Max;
        private final byte[] b2cIsSupp = mapping.b2cIsSupp;
        private final int dbSegSize = b2Max - b2Min + 1;
        private final byte[] cnspToIndex = new byte[0x100];

        protected Decoder(FastCharset cs, EUC_TWMapping mapping) {
            super(cs, mapping, 2.0f, 2.0f);

            // adjust from cns planeNo to the plane index of b2c
            Arrays.fill(cnspToIndex, (byte)-1);
            cnspToIndex[0xa2] = 1; cnspToIndex[0xa3] = 2; cnspToIndex[0xa4] = 3;
            cnspToIndex[0xa5] = 4; cnspToIndex[0xa6] = 5; cnspToIndex[0xa7] = 6;
            cnspToIndex[0xaf] = 7;
        }

        protected CoderResult decodeArrayLoop(
                byte[] sa, int sp, int sl, char[] da, int dp, int dl) {

            while (sp < sl) {
//                int byte1 = sa[sp++] & 0xff;
//                if (byte1 < 0x80) {             // ASCII      G0
                byte byte1 = sa[sp++];
                if (byte1 >= 0) {               // ASCII      G0
                    if (dp == dl)
                       return overflow(sp-1, dp);
//                    da[dp++] = (char)byte1;
                    da[dp++] = (char)(byte1 & 0xff);
                } else if (byte1 != SS2) {      // Codeset 1  G1
                    if (sp == sl)
                        return underflow(sp-1, dp);
//                    int size = decode(byte1, sa[sp++] & 0xff, 0);
                    int size = decode(byte1, sa[sp++], 0);
                    if (size <= 0)
                        if (size < 0)
                            return malformed1(sp-2, dp);
                        else
                            return unmappable2(sp-2, dp);
                    if (dp == dl)
                        return overflow(sp-2, dp);
                    da[dp++] = c1;
                } else {                        // Codeset 2  G2
                    if (sp + 2 >= sl)
                        return underflow(sp-1, dp);
                    int cnsPlane = cnspToIndex[sa[sp++] & 0xff];
                    if (cnsPlane < 0)
                        return malformed2(sp-2, dp);
//                    byte1 = sa[sp++] & 0xff;
//                    int byte2 = sa[sp++] & 0xff;
                    byte1 = sa[sp++];
                    byte byte2 = sa[sp++];
                    int size = decode(byte1, byte2, cnsPlane);
                    if (size <= 0)
                        if (size < 0)
                            return malformed1(sp-4, dp);
                        else
                            return unmappable2(sp-4, dp);
                    if (dp + size > dl)
                        return overflow(sp-4, dp);
                    if (size == 1)
                        da[dp++] = c1;
                    else {
                        da[dp++] = C2[0];
                        da[dp++] = C2[1];
                    }
                }
            }
            return underflow(sp-0, dp);
        }

        protected CoderResult decodeBufferLoop(ByteBuffer src, CharBuffer dst) {

            while (src.hasRemaining()) {
//                int byte1 = src.get() & 0xff;
//                if (byte1 < 0x80) {             // ASCII      G0
                byte byte1 = src.get();
                if (byte1 >= 0) {               // ASCII      G0
                    if (!dst.hasRemaining())
                       return overflow(-1);
//                    dst.put((char)byte1);
                    dst.put((char)(byte1 & 0xff));
                } else if (byte1 != SS2) {      // Codeset 1  G1
                    if (!src.hasRemaining())
                        return underflow(-1);
//                    int byte2 = src.get() & 0xff;
                    byte byte2 = src.get();
                    int size = decode(byte1, byte2, 0);
                    if (size <= 0)
                        if (size < 0)
                            return malformed1(-2);
                        else
                            return unmappable2(-2);
                    if (!dst.hasRemaining())
                        return overflow(-2);
                    dst.put(c1);
                } else {                        // Codeset 2  G2
                    if (3 > src.remaining())
                        return underflow(-1);
                    int cnsPlane = cnspToIndex[src.get() & 0xff];
                    if (cnsPlane < 0)
                        return malformed2(-2);
//                    byte1 = src.get() & 0xff;
//                    int byte2 = src.get() & 0xff;
                    byte1 = src.get();
                    byte byte2 = src.get();
                    int size = decode(byte1, byte2, cnsPlane);
                    if (size <= 0)
                        if (size < 0)
                            return malformed1(-4);
                        else
                            return unmappable2(-4);
                    if (size > dst.remaining())
                        return overflow(-4);
                    if (size == 1)
                        dst.put(c1);
                    else
                        dst.put(C2);
                }
            }
            return CoderResult.UNDERFLOW;
        }

        volatile char c1, c2;
//        private int decode(int b1, int b2, int p) { // TODO: little faster by returning boolean
        int decode(byte b1, byte b2, int p) { // TODO: little faster by returning boolean
            assert b1Max - b1Min >= 0 && b2Max - b2Min >= 0; // important if values are of type byte
            if (b1 < b1Min || b1 > b1Max || b2 < b2Min || b2 > b2Max)
                return -1;
            int index = (b1 - b1Min) * dbSegSize + b2 - b2Min;
            c1 = b2c[p].charAt(index);
            if (c1 == UNMAPPABLE_DECODING)
                return 0;
            if ((b2cIsSupp[index] & (1 << p)) == 0)
                return 1;
            else
                return Character.toChars(0x20000 + c1, C2, 0); // returns always 2
//                return Character.toChars(0x20000 + c1, dst, dp); // TODO: faster
        }

        private char repl = REPLACE_CHAR; // TODO: better refer to default of Charset class?
        protected void implReplaceWith(String newReplacement) {
            repl = newReplacement.charAt(0);
        }

        // Attention: this method is not thread-safe (c1 + C2 are instance members)
        public int decode(byte[] src, int sp, int len, char[] dst) {
            if (len > dst.length)
                len = dst.length;
            int dp = 0;
            while (dp < len) {
//                int byte1 = src[sp++] & 0xff;
                byte byte1 = src[sp++];
//                if (byte1 < 0x80)                   // ASCII      G0
                if (byte1 >= 0)                     // ASCII      G0
//                    dst[dp++] = (char)byte1;
                    dst[dp++] = (char)(byte1 & 0xff);
                else if (byte1 != SS2)              // Codeset 1  G1
//                    dst[dp++] = decode(byte1, src[sp++] & 0xff, 0) == 0 ? repl : c1;
                    dst[dp++] = decode(byte1, src[sp++], 0) == 0 ? repl : c1;
                else {                              // Codeset 2  G2
                    int cnsPlane = cnspToIndex[src[sp++] & 0xff];
//                    int size = decode(src[sp++] & 0xff, src[sp++] & 0xff, cnsPlane);
                    int size = decode(src[sp++], src[sp++], cnsPlane);
                    if (cnsPlane < 0 || size == 0)
                        dst[dp++] = repl;
                    else if (size == 1)
                        dst[dp++] = c1;
                    else if (dp == len)
                        return --dp;
                    else {
                        dst[dp++] = C2[0];
                        dst[dp++] = C2[1];
                    }
                }
            }
            return dp;
        }

        private final char[] C1 = new char[1], C2 = new char[2];
//        public char[] getUnicode(int b1, int b2, int p) {
        public char[] getUnicode(byte b1, byte b2, int p) {
            switch (decode(b1, b2, p)) {
//            switch (decode((byte)b1, (byte)b2, p)) {
                default: return null;
                case 1: C1[0] = c1; return C1;
                case 2: return C2;
            }
        }
    }


    public static class Encoder extends SoftReferenceCharset.Encoder<EUC_TWMapping> {

        private final byte[] bb = new byte[4];
        private final char[] c2b;
        private final char[] c2bIndex;
        private final char[] c2bSupp;
        private final char[] c2bSuppIndex;
        private final byte[] c2bPlane;

        public Encoder(FastCharset cs, EUC_TWMapping mapping) {
            super(cs, 4.0f, 4.0f);

//            int b1Min        = mapping.b1Min;
//            int b1Max        = mapping.b1Max;
//            int b2Min        = mapping.b2Min;
//            int b2Max        = mapping.b2Max;
            int b1Min        = mapping.b1Min & 0xff;
            int b1Max        = mapping.b1Max & 0xff;
            int b2Min        = mapping.b2Min & 0xff;
            int b2Max        = mapping.b2Max & 0xff;
            String[] b2c     = mapping.b2c;
            byte[] b2cIsSupp = mapping.b2cIsSupp;

            c2bIndex = mapping.c2bIndex;
            c2bSuppIndex = mapping.c2bSuppIndex;
            c2b = new char[mapping.C2BSIZE];
            c2bSupp = new char[mapping.C2BSUPPSIZE];
            c2bPlane = new byte[Math.max(mapping.C2BSIZE, mapping.C2BSUPPSIZE)];

            Arrays.fill(c2b, (char)UNMAPPABLE_ENCODING);
            Arrays.fill(c2bSupp, (char)UNMAPPABLE_ENCODING);

            for (int p = 0; p < b2c.length; p++) {
                String db = b2c[p];
                /*
                   adjust the "plane" from 0..7 to 0, 2, 3, 4, 5, 6, 7, 0xf,
                   which helps balance between footprint (to save the plane
                   info in 4 bits) and runtime performance (to require only
                   one operation "0xa0 | plane" to encode the plane byte)
                */
                int plane = p;
                if (plane == 7)
                    plane = 0xf;
                else if (plane != 0)
                    plane = p + 1;

                int off = 0;
                for (int b1 = b1Min; b1 <= b1Max; b1++) {
                    for (int b2 = b2Min; b2 <= b2Max; b2++) {
                        char c = db.charAt(off);
                        if (c != UNMAPPABLE_DECODING) {
                            if ((b2cIsSupp[off] & (1 << p)) != 0) {
                                int index = c2bSuppIndex[c >> 8] + (c&0xff);
                                c2bSupp[index] = (char)((b1 << 8) + b2);
                                c2bPlane[index] |= (byte)(plane << 4);
                            } else {
                                int index = c2bIndex[c >> 8] + (c&0xff);
                                c2b[index] = (char)((b1 << 8) + b2);
                                c2bPlane[index] |= (byte)plane;
                            }
                        }
                        off++;
                    }
                }
            }
        }

        public boolean canEncode(char c) {
//            return (c < '\u0080' || encode(c, bb) != 0);
            return (encode(c, bb) != 0);
        }

        public boolean canEncode(CharSequence cs) {
            int i = 0;
            while (i < cs.length()) {
                char c = cs.charAt(i++);
                if (Character.isHighSurrogate(c)) {
                    if (i == cs.length())
                        return false;
                    char low = cs.charAt(i++);
                    if (!Character.isLowSurrogate(low) || encode(c, low, bb) == 0)
                        return false;
                } else if (!canEncode(c))
                    return false;
            }
            return true;
        }

        public int encodeChar(char c, byte[] bb) {
            return encode(c, bb);
        }

        protected CoderResult encodeArrayLoop(
                char[] sa, int sp, int sl, byte[] da, int dp, int dl) {

            int outSize;
            while (sp < sl) {
                char c = sa[sp++];
//                if (c < '\u0080') {     // ASCII inlined ?
//                    bb[0] = (byte)c;
//                    outSize = 1;
//                } else
EncodeChar:     {
//                    if ((outSize = encode(c, bb)) == -1) {
                    if ((outSize = encode(c, bb)) == 0) {
                        // To check surrogates only after BMP failed
                        // has the benefit of improving the BMP encoding
                        // 10% faster, with the price of the slowdown of
                        // supplementary character encoding.  Given the use
                        // of supplementary characters is really rare, this
                        // is something worth doing.
                        if (Character.isHighSurrogate(c)) {
                            if (sp == sl)
                                return underflow(sp-1, dp);
                            if (!Character.isLowSurrogate(sa[sp]))
                                return malformed1(sp-1, dp);
                            if ((outSize = encode(c, sa[sp++], bb)) == 0)
                                return unmappable2(sp-2, dp);
                            if (dp + outSize > dl)
                                return overflow(sp-2, dp);
                            break EncodeChar;
                        } else if (Character.isLowSurrogate(c))
                            return malformed1(sp-1, dp);
                        else
                            return unmappable1(sp-1, dp);
                    }
                    if (dp + outSize > dl)
                        return overflow(sp-1, dp);
                }
//                while (outSize != 0)
//                    da[dp++] = bb[--outSize];
//                for (int i = outSize-1; i >= 0; i--)
//                    da[dp++] = bb[i];
//                for (int i = outSize; i > 0; )
//                    da[dp++] = bb[--i];
                for (int i = 0; i < outSize; i++)
                    da[dp++] = bb[i];
            }
            return underflow(sp-0, dp);
        }

        protected CoderResult encodeBufferLoop(CharBuffer src, ByteBuffer dst) {

            int outSize;
            while (src.hasRemaining()) {
                char c = src.get();
//                if (c < '\u0080') {     // ASCII inlined ?
//                    bb[0] = (byte)c;
//                    outSize = 1;
//                } else
EncodeChar:     {
                    if ((outSize = encode(c, bb)) == 0) {
                        if (Character.isHighSurrogate(c)) {
                            if (!src.hasRemaining())
                                return underflow(-1);
                            char c2 = src.get();
                            if (!Character.isLowSurrogate(c2))
                                return malformed1(-2);
                            if ((outSize = encode(c, c2, bb)) == 0)
                                return unmappable2(-2);
                            if (outSize > dst.remaining())
                                return overflow(-2);
                            break EncodeChar;
                        } else if (Character.isLowSurrogate(c))
                            return malformed1(-1);
                        else
                            return unmappable1(-1);
                    }
                    if (outSize > dst.remaining())
                        return overflow(-1);
                }
                while (outSize != 0)
                    dst.put(bb[--outSize]);
            }
            return CoderResult.UNDERFLOW;
        }

        private int encode(char c, byte[] bb) {
            if (c < '\u0080') {     // ASCII
                bb[0] = (byte)c;
                return 1;
            }
            int index = c2bIndex[c >> 8];
            if (index  == UNMAPPABLE_ENCODING)
//                return -1;
                return 0;
            index = index + (c & 0xff);
            int db = c2b[index];
            if (db == UNMAPPABLE_ENCODING)
//                return -1;
                return 0;
            int p = c2bPlane[index] & 0xf;
            if (p == 0) {
                bb[0] = (byte)(db >> 8);
                bb[1] = (byte)db;
//                bb[0] = (byte)db;
//                bb[1] = (byte)(db >> 8);
                return 2;
            } else {
                bb[0] = (byte)SS2;
                bb[1] = (byte)(0xa0 | p);
                bb[2] = (byte)(db >> 8);
                bb[3] = (byte)db;
//                bb[0] = (byte)db;
//                bb[1] = (byte)(db >> 8);
//                bb[2] = (byte)(0xa0 | p);
////                bb[3] = (byte)SS2;
//                bb[3] = SS2;
                return 4;
            }
        }

        private int encode(char hi, char low, byte[] bb) {
            int c = Character.toCodePoint(hi, low) - 0x20000;
            if (c < 0 || c > 0xffff)
                return 0;
            int index = c2bSuppIndex[c >> 8];
            if (index  == UNMAPPABLE_ENCODING)
                return 0;
            index = index + (c & 0xff);
            int db = c2bSupp[index];
            if (db == UNMAPPABLE_ENCODING)
                return 0;
            int p = (c2bPlane[index] >> 4) & 0xf;
            bb[0] = (byte)db;
            bb[1] = (byte)(db >> 8);
            bb[2] = (byte)(0xa0 | p);
//            bb[3] = (byte)SS2;
            bb[3] = SS2;
            return 4;
        }

        private byte repl = (byte)'?'; // TODO: better refer to default of Charset class?
        protected void implReplaceWith(byte[] newReplacement) {
            repl = newReplacement[0]; // TODO: what's about replacements with size > 1 ?
        }

        // Attention: this method is not thread-safe (bb[0..3] are instance members)
        public int encode(char[] src, int sp, int len, byte[] dst) {
            int dp = 0;
            int sl = sp + Math.min(len, dst.length);
            int outSize;

            while (sp < sl) {
                char c = src[sp++];
//                if (c < '\u0080') {     // ASCII inlined ?
//                    bb[0] = (byte)c;
//                    outSize = 1;
//                } else
EncodeChar:     if ((outSize = encode(c, bb)) == 0) {
//                if ((outSize = encode(c, dst, dp)) == 0) { // TODO: faster
                    if (Character.isHighSurrogate(c) && sp < sl &&
                            Character.isLowSurrogate(src[sp])) {
                        if (len > dst.length) {
                            sl++;
                            len--;
                        }
                        if ((outSize = encode(c, src[sp++], bb)) != 0)
//                        if ((outSize = encode(c, src[sp++], dst, dp)) != 0) // TODO: faster
                            break EncodeChar;
                    }
                    dst[dp++] = repl;
                    continue;
                }
                if (len - dp < outSize)
                    return dp;
                while (outSize != 0)
                    dst[dp++] = bb[--outSize];
            }
            return dp;
        }
    }
}
