/*
 * @(#)ExtendedCharsets.java        1.44        06/03/20
 *
 * Copyright 2006 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 * Contributors: Ulf Zibis, Cologne <Ulf.Zibis at CoSoCo.de>
 */

package sun.nio.cs.ext;

import java.io.IOException;
import java.nio.charset.Charset;
import java.security.AccessController;
import java.util.List;
import java.util.Set;
import sun.security.action.GetPropertyAction;

/**
 * Provider for extended charsets.
 * Regularly instantiated via Class.forName("sun.nio.cs.ext.ExtendedCharsets").newInstance().
 */
public class ExtendedCharsets extends sun.nio.cs.ExternalCharsetProvider {

    private static volatile ExtendedCharsets instance = null;

    private boolean customized = false;

    public ExtendedCharsets() throws IOException {
        super(new InitData(ExtendedCharsets.class));
        instance = this;
    }

    // If the sun.nio.cs.map property is defined on the command line we won't
    // see it in the system-properties table until after the charset subsystem
    // has been initialized.  We therefore delay the effect of this property
    // until after the JRE has completely booted.
    //
    // At the moment following values for this property are supported, property
    // value string is case insensitive.
    //
    // (1)"Windows-31J/Shift_JIS"
    // In 1.4.1 we added a correct implementation of the Shift_JIS charset
    // but in previous releases this charset name had been treated as an alias
    // for Windows-31J, aka MS932. Users who have existing code that depends
    // upon this alias can restore the previous behavior by defining this
    // property to have this value.
    //
    // (2)"x-windows-50221/ISO-2022-JP"
    //    "x-windows-50220/ISO-2022-JP"
    //    "x-windows-iso2022jp/ISO-2022-JP"
    // The charset ISO-2022-JP is a "standard based" implementation by default,
    // which supports ASCII, JIS_X_0201 and JIS_X_0208 mappings based encoding
    // and decoding only.
    // There are three Microsoft iso-2022-jp variants, namely x-windows-50220,
    // x-windows-50221 and x-windows-iso2022jp which behaves "slightly" differently
    // compared to the "standard based" implementation. See ISO2022_JP.java for
    // detailed description. Users who prefer the behavior of MS iso-2022-jp
    // variants should use these names explicitly instead of using "ISO-2022-JP"
    // and its aliases. However for those who need the ISO-2022-JP charset behaves
    // exactly the same as MS variants do, above properties can be defined to
    // switch.
    //
    // If we need to define other charset-alias mappings in the future then
    // this property could be further extended, the general idea being that its
    // value should be of the form
    //
    //   compensating-charset-1/replaceable-charset-1,compensating-charset-2/replaceable-charset-2,...
    //
    // where each charset named to the left of a slash is intended to compensate
    // (most) uses of the charset named to the right of the slash.
    //
    private synchronized void customize() {
        if (customized)
            return;
        charsets.clear();
        if (!sun.misc.VM.isBooted()) // TODO: superfluous ?
            return;

        String map = AccessController.doPrivileged(new GetPropertyAction("sun.nio.cs.map"));
        if (map != null) {
            for (String mapping : map.split(",")) {
                int separator = mapping.indexOf('/');
                if (separator < 1 || mapping.length()-separator < 2)
                    continue; // illegal formed mapping pair
                List<String> replaceableAliases = unregister(mapping.substring(separator+1));
                if (replaceableAliases == null)
                    continue; // 1st charset is unknown
                List<String> compensatingAliases = unregister(mapping.substring(0, separator));
                if (compensatingAliases == null) {
                    register(replaceableAliases.remove(replaceableAliases.size()-1), // re-register 1st charset
                            replaceableAliases.remove(replaceableAliases.size()-1), replaceableAliases.size() == 0 ?
                                (String[])null : replaceableAliases.toArray(new String[replaceableAliases.size()]));
                    continue; // 2nd charset is unknown
                }
                String replaceable = replaceableAliases.remove(replaceableAliases.size()-1);
                String compensating = compensatingAliases.remove(compensatingAliases.size()-1);
                String replaceableCls = replaceableAliases.remove(replaceableAliases.size()-1);
                String compensatingCls = compensatingAliases.remove(compensatingAliases.size()-1);
                // register new compensating charset and replaceable solely for historical access
                if (replaceableAliases.size() == 0)
                    compensatingAliases.add(replaceable); // add canonical name of replaceable charset
                else {
                    if (toUpper(compensating).startsWith("X-WINDOWS-5022") || // compensating charset
                            toUpper(compensating).equals("X-WINDOWS-ISO2022JP")) // has no historicalName,
                        compensatingAliases.add(replaceable); // so forget historical usage
                    else // reregister by it's historical name only (must be in first list position)
                        register(replaceableAliases.set(0, replaceable), // TODO: get historicalName directly, instead from first list position
                                replaceableCls, (String[])null);
                    // add replaceable charset's aliases and register compensating
                    compensatingAliases.addAll(replaceableAliases);
                }
                register(compensating, compensatingCls,
                        compensatingAliases.toArray(new String[compensatingAliases.size()]));
            }
        }
        String osName = AccessController.doPrivileged(new GetPropertyAction("os.name"));
        if ("SunOS".equals(osName) || "Linux".equals(osName)) {
            register("x-COMPOUND_TEXT", "COMPOUND_TEXT",
                    "COMPOUND_TEXT", // JDK historical
                    "x11-compound_text",
                    "x-compound-text"
                    );
        }
        customized = true;
    }

    protected Set<Charset> charsets0() {
        customize();
        return super.charsets0();
    }

    // invoker should be synchronized to this CharsetProvider !
    protected Charset lookup(String uppercasedName) {
        customize();
        return super.lookup(uppercasedName);
    }

//    // TODO: could be removed when all sun.nio.cs.ext charsets where updated accordingly
//    static String[] aliasesFor(String charsetName) {
//        if (instance == null)
//            return null;
//        return instance.aliases(toUpper(charsetName));
//    }

    // short cut to embed helper charsets, avoids caching in Charset class and
    // multiple instantiation of same charset; public for sun.awt.motif.X11JIS0201
    public static Charset forName(String charsetName) {
        if (instance == null)
            return null;
        return instance.charsetForName(charsetName);
    }
}
